package Davis::Derived;

# use Data::Dumper;
use POSIX qw(floor ceil);

use strict;

use Exporter;
use vars qw( @ISA @EXPORT @EXPORT_OK );
@ISA = qw( Exporter );
@EXPORT = qw( dew_point celcius farenheit from_bcd to_bcd
	scale_bar scale_temp scale_rain scale_hum
	mbars
	forecast
	compass_point
	miles2knots
	windspeed_description
	temp_hum_index
	heat_index
	summer_simmer_index
	);


sub compass_point
{
   my $deg = shift;

   return undef unless defined $deg;

   my $compass;

   if ($deg < 11.25) { $compass = 'N' }
   elsif ($deg < 33.75) { $compass = 'NNE' }
   elsif ($deg < 56.25) { $compass = 'NE' }
   elsif ($deg < 78.75) { $compass = 'ENE' }
   elsif ($deg < 101.25) { $compass = 'E' }
   elsif ($deg < 123.75) { $compass = 'ESE' }
   elsif ($deg < 146.25) { $compass = 'SE' }
   elsif ($deg < 168.75) { $compass = 'SSE' }
   elsif ($deg < 191.25) { $compass = 'S' }
   elsif ($deg < 213.75) { $compass = 'SSW' }
   elsif ($deg < 236.25) { $compass = 'SW' }
   elsif ($deg < 258.75) { $compass = 'WSW' }
   elsif ($deg < 281.25) { $compass = 'W' }
   elsif ($deg < 303.75) { $compass = 'WNW' }
   elsif ($deg < 326.25) { $compass = 'NW' }
   elsif ($deg < 348.75) { $compass = 'NNW' }
   else { $compass = 'N' }

   return $compass;
}

sub miles2knots
{
   my $miles = shift;
   return $miles / 1.1515;
}

sub windspeed_description
{
   my $mph = shift;

   return undef unless defined $mph;

   my $knots = miles2knots($mph);

   my $desc;

   if ($knots < 1) { $desc = 'Still' }
   elsif ($knots <= 3) { $desc = 'Light Air' }
   elsif ($knots <= 6) { $desc = 'Light Breeze' }
   elsif ($knots <= 10) { $desc = 'Gentle Breeze' }
   elsif ($knots <= 16) { $desc = 'Moderate Breeze' }
   elsif ($knots <= 21) { $desc = 'Fresh Breeze' }
   elsif ($knots <= 27) { $desc = 'Strong Breeze' }
   elsif ($knots <= 33) { $desc = 'Near Gale' }
   elsif ($knots <= 40) { $desc = 'Gale' }
   elsif ($knots <= 47) { $desc = 'Strong Gale' }
   elsif ($knots <= 55) { $desc = 'Storm' }
   elsif ($knots <= 63) { $desc = 'Violent Storm' }
   else { $desc = 'Hurricane' }
}

sub celcius
{
   my $f = shift;
   return ($f - 32) * 5 / 9;
}

sub farenheit
{
   my $c = shift;
   return ($c * 9 / 5) + 32;
}

sub mbars
{
   my $inches = shift;
   return $inches * 1013.2 / 29.92;
}

sub dew_point
{
   my $temp = shift;
   my $hum = shift;

   return undef unless ($hum > 0 && $hum <= 100);
   return undef if ($temp < -100 || $temp > 200);

   my $temp_c = celcius($temp);

   my $ews = ($hum / 100) * exp((17.502*$temp_c)/(240.9+$temp_c));

   my $dew = (240.9 * log($ews)) / (17.5 - log($ews));

   return farenheit($dew);
}

sub from_bcd
{
   my $val = shift;

   10 * int($val / 16) + $val % 16;
}

sub to_bcd
{
   my $val = shift;

   16 * int($val / 10) + $val % 10;
}

# scale_* functions deal with precision, invalid values, and sign

sub scale_temp
{
   my $temp = shift;
   return undef if $temp == 32767;
   $temp -= 65536 if $temp > 32767;

   $temp /= 10;

   # reality check:
   return undef if $temp < -50;
   return undef if $temp > 120;

   return $temp;
}

sub scale_rain
{
   my $rain = shift;
   return $rain / 100;
}

sub scale_bar
{
   my $bar = shift;
   my $correction = shift;
   $bar /= 1000;
   $bar -= $correction if defined $correction;
   return $bar;
}

sub scale_hum
{
   my $hum = shift;
   return undef if ($hum < 0 || $hum > 100);
   return $hum
}

sub forecast
{
   my %params = @_;

   # print STDERR Dumper(\%params);

   my $winddir = $params{winddir} || 'U';
   my $current = sprintf("%.1f", $params{mbars});
   my $previous = sprintf("%.1f", $params{old_mbars});
   my $diff = $current - $previous;
   my $change = ($diff < 0) ? 'F' : 'R';	# rising or falling

   if (abs($diff) < 0.5) { $change = 'Steady'; }
   elsif (abs($diff) < 3) { $change .= 'S'; }	# slowly
   elsif (abs($diff) < 6) { $change .= 'M'; }	# moderately
   else { $change .= 'R'; }	# rapidly

   # warn "Change is $change";

   my $result = '';

   if ( $current >= 1019 && $current <= 1023 )
   {
      # warn "In A";
      if ( $change eq 'Steady' )	# barometer steady
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are SW-NW)
little change next 1-2 days
";
         }
         else
         {
            if (225 <= $winddir && $winddir <= 315)
            {
               $result .= "little change next 1-2 days\n\n";
            }
         }
      }
      elsif ( $change eq 'FS' )	# slowly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from S-SE)
rain in 24 hours
(if winds are from SE-E-NE)
increasing wind
rain in 12-18 hours
";
         }
         else
         {
            if (135 <= $winddir && $winddir <= 180)
            {
               $result .= "rain in 24 hours\n\n";
            }
            if (45 <= $winddir && $winddir <= 135)
            {
               $result .= "increasing wind\nrain in 12-18 hours\n\n";
            }
         }
      }
      elsif ( $change eq 'FM' || $change eq 'FR' )	# moderately or rapidly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from S-SE)
rain in 12-24 hours, high wind
(if winds are from SE-E-NE)
increasing wind
rain in 12 hours or less
";
         }
         else
         {
            if (135 <= $winddir && $winddir <= 180)
            {
               $result .= "rain in 12-24 hours, high wind\n\n";
            }
            if (45 <= $winddir && $winddir <= 135)
            {
               $result .= "increasing wind\nrain in 12 hours or less\n\n";
            }
         }
      }
      elsif ( $change eq 'RS' )	# slowly
      {
            # $result .= "Unknown A\n";
      }
      elsif ( $change eq 'RM' )	# moderately
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are SW-NW)
fair
possible rain in 1-2 days
strong winds, rapidly decreasing
";
         }
         else
         {
            if (225 <= $winddir && $winddir <= 315)
            {
               $result .= "rain in 12-24 hours, high wind\n\n";
            }
         }
      }
      elsif ( $change eq 'RR' )	# rapidly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are SW-NW)
fair
possible rain in 1-2 days
strong winds, rapidly decreasing
";
         }
         else
         {
            if (225 <= $winddir && $winddir <= 315)
            {
               $result .= "fair\npossible rain in 1-2 days\nstrong winds, rapidly decreasing\n\n";
            }
         }
      }
      else
      {
         # warn "Bottomed out 1";
      }
   }

   if ( $current <= 1016 )
   {
      # warn "In B";
      if ( $change eq 'Steady' )	# barometer steady
      {
            # $result .= "Unknown B\n";
      }
      elsif ( $change eq 'FS' )	# slowly
      {
         $result .= "(if winds are from SE-E-NE)
rain continuing for 1-3 days
";
      }
      elsif ( $change eq 'FM' || $change eq 'FR' )	# moderately or rapidly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from SE-E-NE)
rain & high wind in 1-6 hours
clearing in 36 hours
colder
";
         }
         else
         {
            if (45 <= $winddir && $winddir <= 135)
            {
               $result .= "rain & high wind in 1-6 hours\nclearing in 36 hours\ncolder\n\n";
            }
         }
      }
      elsif ( $change eq 'RS' )	# slowly
      {
         # $result .= "Unknown C\n";
      }
      elsif ( $change eq 'RM' )	# moderately
      {
         # $result .= "Unknown D\n";
      }
      elsif ( $change eq 'RR' )	# rapidly
      {
         # $result .= "Unknown E\n";
      }
      else
      {
         # warn "Bottomed out 2";
      }
   }

   if ( $current >= 1019 )
   {
      # warn "In C";
      if ( $change eq 'Steady' )	# barometer steady
      {
            # $result .= "Unknown F\n";
      }
      elsif ( $change eq 'FS' )	# slowly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from E-NE)
summer: light wind & rain in 1-3 days
winter: rain & wind in 24 hours
";
         }
         else
         {
            if (45 <= $winddir && $winddir <= 90)
            {
               $result .= "summer: light wind & rain in 1-3 days\nwinter: rain & wind in 24 hours\n\n";
            }
         }
      }
      elsif ( $change eq 'FM' || $change eq 'FR' )	# moderately or rapidly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from E-NE)
summer: rain & wind in 12-24 hours
winter: rain/snow & wind in 12 hours
";
         }
         else
         {
            if (45 <= $winddir && $winddir <= 90)
            {
               $result .= "summer: rain & wind in 12-24 hours\nwinter: rain/show & wind in 12 hours\n\n";
            }
         }
      }
      elsif ( $change eq 'RS' )	# slowly
      {
         # $result .= "Unknown G\n";
      }
      elsif ( $change eq 'RM' )	# moderately
      {
         # $result .= "Unknown H\n";
      }
      elsif ( $change eq 'RR' )	# rapidly
      {
         # $result .= "Unknown I\n";
      }
      else
      {
         # warn "Bottomed out 3";
      }
   }

   if ( $current <= 1009 )
   {
      # warn "In D";
      if ( $change eq 'Steady' )	# barometer steady
      {
            # $result .= "Unknown J\n";
      }
      elsif ( $change eq 'FS' )	# slowly
      {
         # $result .= "Unknown K\n";
      }
      elsif ( $change eq 'FM' || $change eq 'FR' )	# moderately or rapidly
      {
         if ($winddir eq 'U')
         {
            $result .= "(if winds are from N-E-S)
severe storm in few hours
clear in 24 hours
";
         }
         else
         {
            if (0 <= $winddir && $winddir <= 180)
            {
               $result .= "severe storm in few hours\nclear in 24 hours\n\n";
            }
         }
      }
      elsif ( $change eq 'RS' )	# slowly
      {
         # $result .= "Unknown L\n";
      }
      elsif ( $change eq 'RM' )	# moderately
      {
         $result .= "(winds shift rapidly to SW-NW)
high winds up to 24 hours
cold, thunderstorms, showers, then clearing
";
      }
      elsif ( $change eq 'RR' )	# rapidly
      {
         $result .= "(winds shift rapidly to SW-NW)
high winds up to 24 hours
cold, thunderstorms, showers, then clearing
";
      }
      else
      {
         # warn "Bottomed out 4";
      }
   }

   if ( $current >= 1023 )
   {
      # warn "In E";
      if ( $change eq 'Steady' )	# barometer steady
      {
         if ($winddir eq 'U')
         {
         $result .= "(if winds are SW-NW)
continued fair
";
         }
         else
         {
            if (225 <= $winddir && $winddir <= 315)
            {
               $result .= "continued fair\n";
            }
         }
      }
      elsif ( $current < $params{old_mbars} )	# barometer falling
      {
         my $diff = abs($current - $params{old_mbars});
         if ( $diff < 3 )	# slowly
         {
            # $result .= "Unknown M\n";
         }
         elsif ( $diff < 6 )	# moderately
         {
            # $result .= "Unknown N\n";
         }
         else	# rapidly
         {
            # $result .= "Unknown O\n";
         }
      }
      elsif ( $current > $params{old_mbars} ) 	# barometer rising
      {
         my $diff = abs($current - $params{old_mbars});
         if ( $diff < 3 )	# slowly
         {
            # $result .= "Unknown P\n";
         }
         elsif ( $diff < 6 )	# moderately
         {
            # $result .= "Unknown Q\n";
         }
         else	# rapidly
         {
            # $result .= "Unknown R\n";
         }
      }
   }

   return $result;
}

my %thitable = (
         # Humidity %
         #  0  10  20  30  40  50  60  70  80  90 100        F
   68 => [ 61, 63, 63, 64, 66, 66, 68, 68, 70, 70, 70], #  68
   69 => [ 63, 64, 65, 65, 67, 67, 69, 69, 71, 71, 72], #  69
   70 => [ 65, 65, 66, 66, 68, 68, 70, 70, 72, 72, 74], #  70
   71 => [ 66, 66, 67, 67, 69, 69, 71, 71, 73, 73, 75], #  71
   72 => [ 67, 67, 68, 69, 70, 71, 72, 72, 74, 74, 76], #  72
   73 => [ 68, 68, 69, 71, 71, 73, 73, 74, 75, 75, 77], #  73
   74 => [ 69, 69, 70, 72, 72, 74, 74, 76, 76, 76, 78], #  74
   75 => [ 70, 71, 71, 73, 73, 75, 75, 77, 77, 78, 79], #  75
   76 => [ 71, 72, 73, 74, 74, 76, 76, 78, 79, 80, 80], #  76
   77 => [ 72, 73, 75, 75, 75, 77, 77, 79, 81, 81, 82], #  77
   78 => [ 74, 74, 76, 76, 77, 78, 79, 80, 82, 83, 84], #  78
   79 => [ 75, 75, 77, 77, 79, 79, 81, 81, 83, 85, 87], #  79
   80 => [ 76, 76, 78, 78, 80, 80, 82, 83, 85, 87, 90], #  80
   81 => [ 77, 77, 79, 79, 81, 81, 83, 85, 87, 89, 93], #  81
   82 => [ 78, 78, 80, 80, 82, 83, 84, 87, 89, 92, 96], #  82
   83 => [ 79, 79, 81, 81, 83, 85, 85, 89, 91, 95, 99], #  83
   84 => [ 79, 80, 81, 82, 84, 86, 87, 91, 94, 98,103], #  84
   85 => [ 80, 81, 81, 83, 85, 87, 89, 93, 97,101,108], #  85
   86 => [ 81, 82, 82, 84, 86, 88, 91, 95, 99,104,113], #  86
   87 => [ 82, 83, 83, 85, 87, 90, 93, 97,102,109,120], #  87
   88 => [ 83, 84, 84, 86, 88, 92, 95, 99,105,114,131], #  88
   89 => [ 84, 84, 85, 87, 90, 94, 97,102,109,120,144], #  89
   90 => [ 84, 85, 86, 89, 92, 95, 99,105,113,128,150], #  90
   91 => [ 84, 86, 87, 91, 93, 96,101,108,118,136,150], #  91
   92 => [ 85, 87, 88, 92, 94, 98,104,112,124,144,150], #  92
   93 => [ 86, 88, 89, 93, 96,100,107,116,130,150,150], #  93
   94 => [ 87, 89, 90, 94, 98,102,110,120,137,150,150], #  94
   95 => [ 88, 90, 91, 95, 99,104,113,124,144,150,150], #  95
   96 => [ 89, 91, 93, 97,101,107,117,128,150,150,150], #  96
   97 => [ 90, 92, 95, 99,103,110,121,132,150,150,150], #  97
   98 => [ 90, 93, 96,100,105,113,125,150,150,150,150], #  98
   99 => [ 90, 94, 97,101,107,116,129,150,150,150,150], #  99
   100 => [ 91, 95, 98,103,110,119,133,150,150,150,150], # 100
   101 => [ 92, 96, 99,105,112,122,137,150,150,150,150], # 101
   102 => [ 93, 97,100,106,114,125,150,150,150,150,150], # 102
   103 => [ 94, 98,102,107,117,128,150,150,150,150,150], # 103
   104 => [ 95, 99,104,109,120,132,150,150,150,150,150], # 104
   105 => [ 95,100,105,111,123,135,150,150,150,150,150], # 105
   106 => [ 95,101,106,113,126,150,150,150,150,150,150], # 106
   107 => [ 96,102,107,115,130,150,150,150,150,150,150], # 107
   108 => [ 97,103,108,117,133,150,150,150,150,150,150], # 108
   109 => [ 98,104,110,119,137,150,150,150,150,150,150], # 109
   110 => [ 99,105,112,122,142,150,150,150,150,150,150], # 110
   111 => [100,106,113,125,150,150,150,150,150,150,150], # 111
   112 => [100,107,115,128,150,150,150,150,150,150,150], # 112
   113 => [100,108,117,131,150,150,150,150,150,150,150], # 113
   114 => [101,109,119,134,150,150,150,150,150,150,150], # 114
   115 => [102,110,121,136,150,150,150,150,150,150,150], # 115
   116 => [103,111,123,140,150,150,150,150,150,150,150], # 116
   117 => [104,112,125,143,150,150,150,150,150,150,150], # 117
   118 => [105,113,127,150,150,150,150,150,150,150,150], # 118
   119 => [106,114,129,150,150,150,150,150,150,150,150], # 119
   120 => [107,116,131,150,150,150,150,150,150,150,150], # 120
   121 => [108,117,133,150,150,150,150,150,150,150,150], # 121
   122 => [108,118,136,150,150,150,150,150,150,150,150]  # 122
);

sub temp_hum_index
{
   my $temp = shift;
   my $hum = shift;
   $hum /= 10;

   return 'LO' if ($temp < 68);
   return 'HI' if ($temp > 122);

   my $tfloor = floor($temp);
   my $tceil = ceil($temp);

   my $hfloor = floor($hum);
   my $hceil = ceil($hum);

   my $i1 = $thitable{$tfloor}->[$hfloor];
   my $i2 = $thitable{$tfloor}->[$hceil];
   my $i3 = $thitable{$tceil}->[$hfloor];
   my $i4 = $thitable{$tceil}->[$hceil];

   my ($i5, $i6);

   if ($hceil == $hfloor)
   {
      $i5 = $i1;
      $i6 = $i3;
   }
   else
   {
      $i5 = ($hum - $hfloor) / ($hceil - $hfloor) * ($i2 - $i1) + $i1;
      $i6 = ($hum - $hfloor) / ($hceil - $hfloor) * ($i4 - $i3) + $i3;
   }

   my $thi;

   if ($tceil == $tfloor)
   {
      $thi = $i5;
   }
   else
   {
      $thi = ($temp - $tfloor) / ($tceil - $tfloor) * ($i6 - $i5) + $i5;
   }

   return $thi;
}

sub heat_index
{
   my $temp = shift;
   my $hum = shift;

   my $HI = -42.379
        + 2.04901523 * $temp
        + 10.14333127 * $hum
        - 0.22475541 * $temp * $hum
        - 6.83783E-3 * $temp**2
        - 5.481717E-2 * $hum**2
        + 1.22874E-3 * $temp**2 * $hum
        + 8.5282E-4 * $temp * $hum**2
        - 1.99E-6 * $temp**2 * $hum**2
        ;

   $HI;
}

sub summer_simmer_index
{
   my $temp = shift;
   my $hum = shift;

   my $SSI = 1.98 * ($temp - (0.55 - 0.0055 * $hum) * ($temp-58)) - 56.83;
}


1;

