package Davis::VantagePro;

use POSIX qw(ceil);
use Carp;

use strict;

use Davis::CRC;
use Davis::DataPoint;

# use Exporter;
# use vars qw( @ISA @EXPORT @EXPORT_OK );
# @ISA = qw( Exporter );
# @EXPORT = qw( dew_point celcius farenheit );

use Device::SerialPort;

my $default_timeout = 1500;

my $printables = join("", map { POSIX::isprint(chr($_)) ? chr($_) : undef } 0..255);

sub new
{
   my $type = shift;

   my $self = bless { @_ };
   bless $self, $type;

   $self->{baud} = 19200;
#   die "Invalid baud rate: " . $self->{baud}
#	unless ($self->{baud} == 1200 || $self->{baud} == 2400);

   $self->{device} ||= '/dev/davis';

   $self->{port} = new Device::SerialPort($self->{device});
   $self->{port}->baudrate($self->{baud});
   $self->{port}->databits(8);
   $self->{port}->parity('none');
   $self->{port}->stopbits(1);
   $self->{port}->handshake('none');
   $self->{port}->stty_echo(0);
   $self->{port}->read_const_time($default_timeout);
   $self->{port}->read_char_time(5);

   # initialize the thing -- can we determine if it's Rev E or not?

   $self->wrd || return undef;	# if it's not really a Vantage
   # $self->lamps(1);

   return $self;
}

## END CONSTRUCTORS


sub dumpstring
{
   my $self = shift;
   my $string = shift;

   my $per_line = 8;

   my $output = '';

   $output .= sprintf "\nLength: %s\n\n0x0000: ", length($string);

   my $offset = 0;

   foreach (split(//, $string))
   {
      $offset++;

      $output .= sprintf "  %3d=%02x=%1s",
		ord($_),
		ord($_),
		( $_ =~ /^[\w]$/i ) ? $_ : ' ',
		;

      unless ($offset % $per_line)
      {
         $output .= sprintf "\n0x%04x: ", $offset;
      }      
   }

   $output .= "\n" if ($offset % $per_line);

   $output .= "\n";

   return $output;
}


sub get_ok
{
   my $self = shift;

   my ($count, $buf) = $self->{port}->read(6);

   return 'true' if $buf eq "\n\rOK\n\r";

   warn "Not OK";
   print STDERR $self->dumpstring($buf);

   return undef;
}

sub getln
{
   my $self = shift;

   my $buf = '';

   LINEBYTE:
   while ($buf !~ /\n\r$/)
   {
      my ($count, $newbuf) = $self->{port}->read(1);

      if ($count)
      {
         $buf .= $newbuf;
      }
      else
      {
         return undef;
      }
   }

   $buf =~ s/[\r\n]+$//;

   return $buf;
}

sub wakeup
{
   my $self = shift;

   $self->{port}->read_const_time(100);

   my $good = 0;
   my $count = 0;

   TRY:
   foreach (1..10)
   {
      $self->{port}->write("\r");

      my ($count, $buf) = $self->{port}->read(2);

      if ($count == 2 && $buf eq "\n\r")
      {
         $good = 1;
         last TRY;
      }
   }

   $self->{port}->read_const_time($default_timeout);

   return 'true' if $good;

   warn "Couldn't wakeup...";
   return undef;
}

sub test
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("TEST\r");

   my ($count, $buf) = $self->{port}->read(8);

   return 'true' if ($count == 8 && $buf eq "\n\rTEST\n\r");

   return undef;
}

sub lamps
{
   my $self = shift;

   my $lamps = shift;

   $lamps = $lamps ? 1 : 0;

   $self->wakeup || return undef;

   $self->{port}->write("LAMPS $lamps\r");

   return $self->get_ok;
}

sub loop
{
   my $self = shift;
   my $loops = shift;

   $loops ||= 1;

   $self->wakeup || return undef;

   $self->{port}->write("LOOP $loops\r");

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read(99);

   return undef unless ( $count == 99 );
   return undef unless ( "\n" eq substr($buf,-4,1) );
   return undef unless ( "\r" eq substr($buf,-3,1) );
   return undef unless ( 0 == davis_crc($buf) );

   # print $self->dumpstring($buf);

   # middle section shows the info we'll keep...
   my $format = 'A4C' . 'vvvCvCCvC15C8vCvvvvvvvvvC4C4CCvC8C4CvCCvv' . 'CCv';
   my @values = unpack($format, $buf);

   shift @values;	# remove 'LOOP'
   shift @values;	# remove \0
   pop @values;		# remove crc
   pop @values;		# remove \r
   pop @values;		# remove \n

   # icon bits:
   # 1 = Rain
   # 2 = Clouds
   # 4 = Partial Sun
   # 8 = Full Sun
   # 16 =
   # 32 =
   # 64 =
   # 128 =

   my %icons = (
	0 => 'None',
	3 => 'Clouds/Rain',
	6 => 'Partial Sun/Clouds',
	7 => 'Partial Sun/Clouds/Rain',
	8 => 'Sunny',
	);

   my %rules = (
	1 => '1 => 8',
	9 => '9 => 6',
	12 => '12 => 6',
	44 => '44 => 6',
	45 => '45 => 6',
	75 => '75 => 6',
	93 => '93 => 7',
	161 => '161 => 3',
	184 => '184 => 3',
	192 => '192 => 3',
	);

   my $data = {
	next_record	=> $values[0],
	barometer	=> $values[1]/1000,
	inside_temp	=> $values[2]/10,
	inside_hum	=> $values[3],
	outside_temp	=> ($values[4] == 32767) ? undef : $values[4]/10,
	wind_speed	=> $values[5],
	wind_speed_10min => $values[6],
	wind_dir	=> $values[7],
	outside_hum	=> $values[23],
	rain_rate	=> $values[31]/100,
	# uv		=> $values[32],
	# solar		=> $values[33],
	rain_storm	=> $values[34]/100,
	storm_start_date => $values[35],
	rain_day	=> $values[36]/100,
	rain_month	=> $values[37]/100,
	rain_year	=> $values[38]/100,
	# et_day	=> $values[39],
	# et_month	=> $values[40],
	# et_year	=> $values[41],
	# alarms_inside	=> $values[50],
	# alarms_rain	=> $values[51],
	# alarms_outside	=> $values[52],
	sunrise		=> sprintf("%s:%02s", int($values[69]/100), $values[69] % 100),
	sunset		=> sprintf("%s:%02s", int($values[70]/100), $values[70] % 100),
	forecast_icons => $icons{$values[67]} || $values[67],
	forecast_rule => $rules{$values[68]} || $values[68],
	};

   # dups/derived
   $data->{rain} = $data->{rain_storm};

   return $data;
}

sub get_next_archive_index
{
   my $self = shift;
   my $loop = $self->loop;
   return $loop->{next_record};
}

sub get_oldest_archive_index
{
   return 0;
}

sub wrd
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("WRD\x12\x4d\r");

   my ($count, $buf) = $self->{port}->read(2);

   die unless ($buf eq "\x06\x10");

   return 'true';
}

sub rxcheck
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("RXCHECK\r");

   return undef unless $self->get_ok;

   my $buf = $self->getln;

   my @list = split(/\s+/, $buf);

   my $data = {
	received_packets	=> $list[0],
	missed_packets		=> $list[1],
	unk2			=> $list[2],
	longest_good_streak	=> $list[3],
	crc_errors		=> $list[4],
	};

   return $data;
}

sub gettime
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("GETTIME\r");

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read(8);

   return undef unless (0 == davis_crc($buf));

   my @list = map { ord($_) } split(//, $buf);

   my $data = {
	second	=> $list[0],
	minute	=> $list[1],
	hour	=> $list[2],
	day	=> $list[3],
	month	=> $list[4],
	year	=> $list[5] + 1900,
	};

   return $data;
}

sub settime
{
   my $self = shift;
   my $second = shift;
   my $minute = shift;
   my $hour = shift;
   my $day = shift;
   my $month = shift;
   my $year = shift;

   $self->wakeup || return undef;

   $self->{port}->write("SETTIME\r");

   my $ack1 = $self->get_ack;
   # first ack only indicates command reception
   return $ack1 unless ($ack1 && $ack1 == 0);

   my $data = pack('C6', $second, $minute, $hour, $day, $month, $year-1900);

   my $crc = pack('n', davis_crc($data));

   die "Ugh" unless 0 == davis_crc($data . $crc);

   $self->{port}->write($data . $crc . "\r");

   # second ack may indicate bad crc
   my $ack2 = $self->get_ack;

   return $ack2 ? $ack2 : 'true';
}

sub get_archive_entry
{
   my $self = shift;
   my $recno = shift;

   $self->wakeup || return undef;

   # $self->buffwr;

   $self->{port}->write("RECRD" . pack('v', $recno) . "\r");

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read(54);

   return undef unless (0 == davis_crc($buf));

   # print $self->dumpstring($buf);

   my @list = unpack('vvvvvvvvvvvCCCCCCCCC4C4C4C2C2vvvv', $buf);

   return undef if $list[0] == 0xffff;

   # printf STDERR "0x%04x\n", $list[0];

   my $date = sprintf("%04s%02s%02s",
	2000 + ( ($list[0] >> 9) & 0x7f ),
	($list[0] >> 5) & 0xf,
	$list[0] & 0x1f,
	);

   my %data = (
	date		=> $date,
	year		=> 2000 + ( ($list[0] >> 9) & 0x7f ),
	month		=> ($list[0] >> 5) & 0xf,
	day		=> $list[0] & 0x1f,
	time		=> $list[1],
	outside_temp	=> ($list[2] == 32767) ? undef : $list[2]/10,
	outside_temp_hi	=> $list[3]/10,
	outside_temp_lo	=> $list[4]/10,
	interval_rain	=> $list[5]/100,
	rain_rate	=> $list[6]/100,
	barometer	=> $list[7]/1000,
	# solar		=> $list[8],
	wind_samples	=> $list[9],
	inside_temp	=> $list[10]/10,
	inside_hum	=> ($list[11] > 100) ? undef : $list[11],
	outside_hum	=> ($list[12] > 100) ? undef : $list[12],
	wind_speed	=> ($list[13] == 255) ? undef : $list[13],
	wind_gust	=> $list[14],
	wind_dir_hi	=> ($list[15] == 255) ? undef : $list[15],
	wind_dir	=> ($list[16] == 255) ? undef : $list[16],
	# uv		=> $list[17],
	# et		=> $list[18],
	# soil_temp	=> $list[19],
	# soil_moist_1	=> $list[20],
	# soil_moist_2	=> $list[21],
	# soil_moist_3	=> $list[22],
	# soil_moist_4	=> $list[23],
	# leaf_temp_1	=> $list[24],
	# leaf_temp_2	=> $list[25],
	# leaf_temp_3	=> $list[26],
	# leaf_temp_4	=> $list[27],
	# leaf_wet_1	=> $list[28],
	# leaf_wet_2	=> $list[29],
	# leaf_wet_3	=> $list[30],
	# leaf_wet_4	=> $list[31],
	# extra_temp_1	=> $list[32],
	# extra_temp_2	=> $list[33],
	# extra_hum_1	=> $list[34],
	# extra_hum_2	=> $list[35],
	# reed_closed	=> $list[36],
	# reed_open	=> $list[37],
	# crc		=> $list[38],
	);

   return new Davis::DataPoint(%data);
}

#sub get_bar_correction
#{
#   return 0;
#}

sub start
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("START\r");

   return undef unless $self->get_ok;

   return 'true';
}

sub stop
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("STOP\r");

   return undef unless $self->get_ok;

   return 'true';
}

sub log
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("LOG\r");

   return $self->get_ok;
}

# DON'T USE THIS!
sub sleep
{
   my $self = shift;

   die "Don't Use SLEEP!";

   $self->wakeup || return undef;

   $self->{port}->write("SLEEP\r");

   return $self->get_ok;
}

sub buffwr
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("BUFFWR\r");

   return $self->get_ack;
}

# gets the whole EEPROM image -- 4k plus CRC
sub getee
{
   my $self = shift;

   $self->wakeup || return undef;

   $self->{port}->write("GETEE\r");

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read(4098);

   return undef unless 0 == davis_crc($buf);

   # return $self->dumpstring($buf);

   return substr($buf, 0, 4096);
}

sub eerd
{
   my $self = shift;
   my $addr = shift;
   my $n = shift;

   return undef unless $addr && $n;

   $self->wakeup || return undef;

   $self->{port}->write(sprintf("EERD %X %X\r", $addr, $n));

   return undef unless $self->get_ok;

   my @list;

   foreach (1..$n)
   {
      my $buf = $self->getln;

      if (defined $buf)
      {
         push @list, $buf;
      }
      else
      {
         return undef;
      }
   }

   return @list;
}

sub eewr
{
   my $self = shift;
   my $addr = shift;
   my $val = shift;

   return undef unless $addr;

   $self->wakeup || return undef;

   $self->{port}->write(sprintf("EEWR %X %X\r", $addr, $val));

   return $self->get_ok;
}

sub eebrd
{
   my $self = shift;
   my $addr = shift;
   my $n = shift;

   return undef unless $addr && $n;

   $self->wakeup || return undef;

   $self->{port}->write(sprintf("EEBRD %X %X\r", $addr, $n));

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read($n + 2);

   return undef unless $count == $n + 2;

   return undef unless 0 == davis_crc($buf);

   return substr($buf, 0, $n);
}

sub eebwr
{
   my $self = shift;
   my $addr = shift;
   my $n = shift;
   my $buf = shift;

   return undef unless $addr;
   return undef unless $n == length($buf);

   $self->wakeup || return undef;

   my $crc = pack('n', davis_crc($buf));

   $self->{port}->write("EEBWR\r");

   return undef unless $self->get_ack;

   $self->{port}->write($buf . $crc . "\r");

   return $self->get_ack;
}

sub pagerd
{
   my $self = shift;
   my $addr = shift;

   $self->wakeup || return undef;

   $self->{port}->write(sprintf("PAGERD %X\r", $addr));

   return undef unless $self->get_ack;

   my ($count, $buf) = $self->{port}->read(267);

   return undef unless 0 == davis_crc($buf);

   # my $blocknum = ord(substr($buf,0,1));

   my $page = substr($buf,1,264);

   return $page;
}

sub setper
{
   my $self = shift;
   my $val = shift;

   return undef unless $val > 0;

   $self->wakeup || return undef;

   $self->{port}->write(sprintf("SETPER %s\r", $val));

   return $self->get_ok;
}

sub archive_array_size
{
   # 512 pages with 5 records each
   return 512 * 5;
}








DESTROY
{
   my $self = shift;
   $self->lamps(0);
   # $self->sleep;
}





#### OLD METHODS -- carried over from WM2 stuff...
#### a couple are still used, but most won't do anything w/ the VP

sub get_station_type
{
   my $self = shift;

   my $stationtype = $self->read_station_memory(0, 0x4d, 1);

   return $stationtype;
}

sub read_archive_memory
{
   my $self = shift;
   my $address = shift;
   my $bytes = shift;

   die "Bad count: $bytes" unless ($bytes > 0);

   return undef unless $self->send_command('SRD' . pack('vv', $address, $bytes - 1) . "\r");

   my $buf = $self->read_block($bytes, 1);

   return substr($buf, 0, $bytes);
}

sub read_link_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;

   die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   return undef unless $self->send_command('RRD' . pack('CCC', $bank, $address, $nybbles-1) . "\r");

   my $bytes = ceil($nybbles / 2);

   my $buf = $self->read_block($bytes);

   return undef unless $buf;

   my $value;

   if ($nybbles == 3 || $nybbles == 4)
   {
      $value = unpack("v", $buf);
      $value &= 0x0fff if $nybbles == 3;
   }
   elsif ($nybbles == 1 || $nybbles == 2)
   {
      $value = unpack("C", $buf);
      $value &= 0x000f if $nybbles == 1;
   }

   return $value;
}

sub read_station_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;

   die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   $bank = [2, 4]->[$bank];

   my $countbank = ($nybbles << 4) | $bank;

   return undef unless $self->send_command('WRD' . pack('CC', $countbank, $address) . "\r");

   my $bytes = ceil($nybbles / 2);

   my $buf = $self->read_block($bytes);

   return undef unless $buf;

   my $value;

   if ($nybbles == 3 || $nybbles == 4)
   {
      $value = unpack("v", $buf);
      $value &= 0x0fff if $nybbles == 3;
   }
   elsif ($nybbles == 1 || $nybbles == 2)
   {
      $value = unpack("C", $buf);
      $value &= 0x000f if $nybbles == 1;
   }

   return $value;
}

sub write_station_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;
   my $value = shift;

   # die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   $bank = [1, 3]->[$bank];

   my $countbank = ($nybbles << 4) | $bank;

   my $command = 'WWR' . pack('CC', $countbank, $address);

   if (length($value) != ceil($nybbles/2))
   {
      warn "Length passed doesn't match length specified";
      warn sprintf("Actual: %s; Spec: %s", length($value), $nybbles);
      return undef
   }

   $command .= $value;

   $command .= "\r";

   return undef unless $self->send_command($command);

   return 'true';
}

sub set_archive_interval
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command('SAP' . pack('C', $minutes) . "\r");

   return 'true';
}

sub stop_old
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("STOP\r");

   return 'true';
}

sub start_old
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("START\r");

   return 'true';
}

sub arc
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("ARC\r");

   return 'true';
}

sub img
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("IMG\r");

   return 'true';
}

sub dbt
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("DBT\r");

   return 'true';
}

sub ebt
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("EBT\r");

   return 'true';
}

sub loop_old
{
   my $self = shift;
   my $format = shift;
   my $count = shift || 1; # don't recommend using this parameter for now...

   warn "Packet format must be provided!" unless $format;

   my $size = length(pack($format));

   return undef unless $self->send_command('LOOP' . pack('v', 65536 - $count) . "\r");

   my $buf = $self->read_block($size, 1, 1);

   return undef unless $buf;

   my @data = unpack($format, $buf);

   return @data;
}

sub send_command
{
   my $self = shift;
   my $command = shift;

   my $count = 0;
   my $ack;

   do
   {
      $count++;

      $self->{port}->write($command);

      $ack = $self->get_ack;

   } until ($ack || $count == 5);

   # return 'true' if $ack;
   # return undef;

   return $ack;
}

sub read_block
{
   my $self = shift;
   my $bytes = shift;
   my $crccheck = shift || 0;	# use to FORCE crc check
   my $skip = shift || 0;	# bytes to skip for crc check

   my ($count, $buf) = $self->{port}->read($bytes);

   unless ($count == $bytes)
   {
      warn "Read timeout -- wanted $bytes, got $count";
      return undef;
   }

   return $buf;
}

sub get_ack
{
   my $self = shift;

   my ($count, $buf) = $self->{port}->read(1);

   unless ($count == 1)
   {
      warn "No ACK";
      return undef;
   }

   my $ord = ord($buf);

   if ($buf eq chr(6))
   {
      return 'zero but true';
   }
   elsif ($ord == 0x21)
   {
      warn "Command not understood";
      return $ord;
   }
   elsif ($buf eq chr(24))
   {
      warn "Bad command checksum";
      return 24;
   }
   else
   {
      warn sprintf("Unknown ACK byte: 0x%02x", ord($buf));
      return undef;
   }

}




1;

