package Davis::WeatherLink;

use POSIX qw(ceil);

use strict;

use Davis::CRC;

# use Exporter;
# use vars qw( @ISA @EXPORT @EXPORT_OK );
# @ISA = qw( Exporter );
# @EXPORT = qw( dew_point celcius farenheit );

use Device::SerialPort;

sub new
{
   my $type = shift;

   my $self = bless { @_ };
   bless $self, $type;

   $self->{baud} ||= 2400;
   die "Invalid baud rate: " . $self->{baud}
	unless ($self->{baud} == 1200 || $self->{baud} == 2400);

   $self->{device} ||= '/dev/davis';

   $self->{port} = new Device::SerialPort($self->{device});
   $self->{port}->baudrate($self->{baud});
   $self->{port}->databits(8);
   $self->{port}->parity('none');
   $self->{port}->stopbits(1);
   $self->{port}->handshake('none');
   $self->{port}->stty_echo(0);
   $self->{port}->read_const_time(2000);
   $self->{port}->read_char_time(5);

   # initialize the thing -- can we determine if it's Rev E or not?

   return $self;
}

## END CONSTRUCTORS

sub get_station_type
{
   my $self = shift;

   my $stationtype = $self->read_station_memory(0, 0x4d, 1);

   return $stationtype;
}

sub read_archive_memory
{
   my $self = shift;
   my $address = shift;
   my $bytes = shift;

   die "Bad count: $bytes" unless ($bytes > 0);

   return undef unless $self->send_command('SRD' . pack('vv', $address, $bytes - 1) . "\r");

   my $buf = $self->read_block($bytes, 1);

   return substr($buf, 0, $bytes);
}

sub read_link_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;

   die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   return undef unless $self->send_command('RRD' . pack('CCC', $bank, $address, $nybbles-1) . "\r");

   my $bytes = ceil($nybbles / 2);

   my $buf = $self->read_block($bytes);

   return undef unless $buf;

   my $value;

   if ($nybbles == 3 || $nybbles == 4)
   {
      $value = unpack("v", $buf);
      $value &= 0x0fff if $nybbles == 3;
   }
   elsif ($nybbles == 1 || $nybbles == 2)
   {
      $value = unpack("C", $buf);
      $value &= 0x000f if $nybbles == 1;
   }

   return $value;
}

sub read_station_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;

   die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   $bank = [2, 4]->[$bank];

   my $countbank = ($nybbles << 4) | $bank;

   return undef unless $self->send_command('WRD' . pack('CC', $countbank, $address) . "\r");

   my $bytes = ceil($nybbles / 2);

   my $buf = $self->read_block($bytes);

   return undef unless $buf;

   my $value;

   if ($nybbles == 3 || $nybbles == 4)
   {
      $value = unpack("v", $buf);
      $value &= 0x0fff if $nybbles == 3;
   }
   elsif ($nybbles == 1 || $nybbles == 2)
   {
      $value = unpack("C", $buf);
      $value &= 0x000f if $nybbles == 1;
   }

   return $value;
}

sub write_station_memory
{
   my $self = shift;
   my $bank = shift;
   my $address = shift;
   my $nybbles = shift;
   my $value = shift;

   # die "Bad count: $nybbles" unless ($nybbles > 0 && $nybbles <= 4);

   $bank = [1, 3]->[$bank];

   my $countbank = ($nybbles << 4) | $bank;

   my $command = 'WWR' . pack('CC', $countbank, $address);

   if (length($value) != ceil($nybbles/2))
   {
      warn "Length passed doesn't match length specified";
      warn sprintf("Actual: %s; Spec: %s", length($value), $nybbles);
      return undef
   }

   $command .= $value;

#   if ($nybbles == 3 || $nybbles == 4)
#   {
#      $command .= pack("v", $value);
#   }
#   elsif ($nybbles == 1 || $nybbles == 2)
#   {
#      $command .= pack("C", $value);
#   }
#   else
#   {
#      warn "write_station_memory: unsupported length";
#      return undef;
#   }

   $command .= "\r";

   return undef unless $self->send_command($command);

   return 'true';
}

sub set_archive_interval
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command('SAP' . pack('C', $minutes) . "\r");

   return 'true';
}

sub stop
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("STOP\r");

   return 'true';
}

sub start
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("START\r");

   return 'true';
}

sub arc
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("ARC\r");

   return 'true';
}

sub img
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("IMG\r");

   return 'true';
}

sub dbt
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("DBT\r");

   return 'true';
}

sub ebt
{
   my $self = shift;
   my $minutes = shift;

   return undef unless $self->send_command("EBT\r");

   return 'true';
}

sub loop
{
   my $self = shift;
   my $format = shift;
   my $count = shift || 1; # don't recommend using this parameter for now...

   warn "Packet format must be provided!" unless $format;

   my $size = length(pack($format));

   return undef unless $self->send_command('LOOP' . pack('v', 65536 - $count) . "\r");

   my $buf = $self->read_block($size, 1, 1);

   return undef unless $buf;

   my @data = unpack($format, $buf);

   return @data;
}

sub send_command
{
   my $self = shift;
   my $command = shift;

   my $count = 0;
   my $ack;

   do
   {
      $count++;

      if ($self->{rev} eq 'E')
      {
         $self->{port}->write(pack('v', davis_davis_crc($command)));
      }

      $self->{port}->write($command);

      $ack = $self->get_ack;

   } until ($ack || $count == 5);

   # return 'true' if $ack;
   # return undef;

   return $ack;
}

sub read_block
{
   my $self = shift;
   my $bytes = shift;
   my $crccheck = shift || 0;	# use to FORCE crc check
   my $skip = shift || 0;	# bytes to skip for crc check

   my ($count, $buf) = $self->{port}->read($bytes);

   unless ($count == $bytes)
   {
      warn "Read timeout -- wanted $bytes, got $count";
      return undef;
   }

   if ($crccheck || $self->{rev} eq 'E')
   {
      my ($crcbytes, $crc) = $self->{port}->read(2);

      unless ($crcbytes == 2)
      {
         warn "Read timeout on CRC";
         return undef;
      }

      return undef unless (davis_crc(substr($buf . $crc, $skip)) == 0);
   }

   return $buf;
}

sub get_ack
{
   my $self = shift;

   my ($count, $buf) = $self->{port}->read(1);

   unless ($count == 1)
   {
      warn "No ACK";
      return undef;
   }

   if ($buf eq chr(6))
   {
      return 'zero but true';
   }
   elsif ($buf eq chr(33))
   {
      warn "Command not understood";
      return 33;
   }
   elsif ($buf eq chr(24))
   {
      warn "Bad command checksum";
      return 24;
   }
   else
   {
      warn sprintf("Unknown ACK byte: 0x%02x", ord($buf));
      return undef;
   }

}


1;

