package Davis::WeatherMonitorII;

use strict;

use vars qw( @ISA );
@ISA = qw( Davis::Console );

use Davis::Console;
use Davis::DataPoint;
use Davis::Derived;

sub new
{
   my $type = shift;

   my $self = new Davis::Console(@_);

   return undef unless $self;

   bless $self, $type;

   my $stationtype = $self->{link}->get_station_type;

   die "Station is not a Weather Monitor II" unless ($stationtype == 2);

   return $self;
}


sub loop
{
   my $self = shift;

   my (
	$sob,		# unused
	$temp_i,
	$temp_o,
	$wind_speed,
	$wind_dir,
	$baro,
	$hum_i,
	$hum_o,
	$rain,
	$junk,		# unused
	) = $self->{link}->loop('CvvCvvCCvv');

   my %data = (
	inside_temp => $temp_i / 10,
	wind_speed => $wind_speed,
	barometer => scale_bar($baro, $self->get_bar_correction),
	inside_hum => $hum_i,
	rain => $rain / 100,
	);

   if ($wind_dir != 32768)
   {
      $data{wind_dir} = $wind_dir;
   }

   if ($temp_o != 32767)
   {
      $data{outside_temp} = $temp_o / 10;
   }

   if ($hum_o != 101)
   {
      $data{outside_hum} = $hum_o;
   }

   return new Davis::DataPoint(%data);

}

sub get_archive_entry
{
   my $self = shift;
   my $entry = shift || 0;

   my $entryformat = 'vCCvvvCCvCCCCCv';
   my $entrysize = length(pack($entryformat));

   my $addr = $self->archive_entry_size * $entry;

   my $buf = $self->{link}->read_archive_memory($addr, $entrysize);

   my (
	$bar,
	$hum_i,
	$hum_o,
	$rain_int,	# interval -- not total
	$temp_i,	# avg over period
	$temp_o,	# avg over period
	$wind_speed,	# avg over period
	$wind_dir,	# dominant over period (0-15)
	$temp_o_hi,
	$wind_gust,
	$hr,
	$min,
	$day,
	$month,
	$temp_o_lo,
	) = unpack($entryformat, $buf);

   # fix precision

   if ($wind_dir == 255)
   {
      $wind_dir = undef;
   }
   else
   {
      $wind_dir *= 22.5;
   }

   if ($hum_o < 0 || $hum_o > 100)
   {
      $hum_o = undef;
   }


   # fix data info

   # warn sprintf("Got month 0x%02x, day 0x%02x", $month, $day);

   my $time = sprintf("%02s%02s", from_bcd($hr), from_bcd($min));
   my $date = sprintf("%02s%02s", $month & 0x0f, from_bcd($day));

   # warn sprintf("Got date %02s, day %02s", $date, $time);

   my %data = (
	date => $date,
	time => $time,
	barometer => scale_bar($bar, $self->get_bar_correction),
	inside_hum => scale_hum($hum_i),
	inside_temp => scale_temp($temp_i),
	outside_hum => scale_hum($hum_o),
	outside_temp => scale_temp($temp_o),
	interval_rain => scale_rain($rain_int),
	wind_dir => $wind_dir,
	wind_speed => $wind_speed,
	wind_gust => $wind_gust,
	outside_temp_lo => $temp_o_lo,
	outside_temp_hi => $temp_o_hi,
	);

   return new Davis::DataPoint(%data);
}

sub archive_array_size
{
   my $self = shift;
   return int($self->archive_memory_size / $self->archive_entry_size);
}

sub archive_entry_size
{
   return 21;
}

sub archive_memory_size
{
   return 32768;
}

sub get_oldest_archive_index
{
   my $self = shift;

   my $val = $self->{link}->read_link_memory(1, 0x04, 4);

   return $val / $self->archive_entry_size;
}

sub get_next_archive_index
{
   my $self = shift;

   my $val = $self->{link}->read_link_memory(1, 0x00, 4);

   return $val / $self->archive_entry_size;
}

sub get_bar_correction
{
   my $self = shift;
   my $attrname = 'bar_correction';

   unless (defined $self->{$attrname})
   {
      my $value;

      while (! defined $value)
      {
         $value = $self->{link}->read_station_memory(1, 0x2c, 4);
      }

      $value -= 65536 if $value > 32767;
      $value /= 1000;

      $self->{$attrname} = $value;
   }

   return $self->{$attrname};
}



1;

